#!/usr/bin/python3
import hashlib
import random
import base64
import time
import json
import asyncio
from client_handler import ClientHandler

RUNNING = False
SESSIONS = []
TASKGROUP = None

class Conversation:
    def __init__(self, addr, info):
        self.info = info
        self.patchnoted = False
        self.user_info = {}
        self.errored = False
        self.host, self.port = addr

    async def run(self):
        if not (await self.init_connection(self.info)):
            return False

        if(self.user_info == {}):
            await asyncio.gather(self.connectivity_checks())
        elif(self.user_info["activated"] == False):
            # Can't ask for patchnotes if account isn't activated
            await asyncio.gather(self.connectivity_checks(),
                                 self.action_loop(self.list_user_parameters))
        else:
            # Don't ask for user parameters if account is activated: would leak the activation key!
            await asyncio.gather(self.connectivity_checks(),
                                 self.action_loop(self.last_patchnotes))

    async def establish_connection(self, host, port):
        for fails in range(5):
            try:
                reader, writer = await asyncio.open_connection(host, port)
                return (reader, writer)
            except:
                await asyncio.sleep(5)

        print("Client unable to connect after one minute", flush=True)
        self.errored = True
        return (None, None)

    async def init_connection(self, info):
        reader, writer = await self.establish_connection(self.host, self.port)
        if not reader and not writer:
            return False

        handler = ClientHandler(reader, writer, SESSIONS)

        if not (await handler.make_request("Hello", 1)):
            writer.close()
            await writer.wait_closed()
            return False
        self.session_id = handler.session_id

        if not (await handler.make_request("version", 5)):
            writer.close()
            await writer.wait_closed()
            return False

        if("username" in info) & ("password" in info):
            b64username = base64.b64encode(info['username'].encode())
            b64password = base64.b64encode(info['password'].encode())
            hashed = hashlib.sha256(b64username + b':' + b64password).hexdigest()

            if not (await handler.make_request(hashed, 2)):
                writer.close()
                await writer.wait_closed()
                return False
            self.user_info = handler.json_data

        writer.close()
        await writer.wait_closed()
        return True

    async def list_user_parameters(self, handler):
        parameters = {"username":"", "password":"", "activation key":""}
        await handler.make_request(json.dumps(parameters), 7)

    async def last_patchnotes(self, handler):
        last_time = int(time.time())
        if not self.patchnoted:
            last_time = 0
            self.patchnoted = True
        await handler.make_request(json.dumps({"ts":last_time}), 9)
                       
    async def action_loop(self, action_function):
        max_delay = 300
        while not self.errored:
            first_wait = random.randrange(3 * max_delay // 5)
            await asyncio.sleep(max_delay // 5 + first_wait)
            reader, writer = await self.establish_connection(self.host, self.port)
            if not reader and not writer:
                return

            handler = ClientHandler(reader, writer, SESSIONS)
            handler.session_id = self.session_id
            await action_function(handler)

            writer.close()
            await writer.wait_closed()

            await asyncio.sleep(4 * max_delay // 5 - first_wait)
                       
    async def connectivity_checks(self):
        consecutive_failed_checks = 0
        total_checks = 0
        max_checks = 10 + random.randrange(10)

        while not self.errored:
            reader, writer = await self.establish_connection(self.host, self.port)
            if not reader and not writer:
                return False

            handler = ClientHandler(reader, writer, SESSIONS)
            handler.session_id = self.session_id
            if not (await handler.make_request(str(int(time.time())), 6)):
                consecutive_failed_checks += 1
                if(consecutive_failed_checks >= 5):
                    self.errored = True
                    return
            else:
                consecutive_failed_checks = 0

            writer.close()
            await writer.wait_closed()

            total_checks += 1
            if total_checks > max_checks:
                self.errored = True
                return

            await asyncio.sleep(60)

async def kill():
    global RUNNING
    tasks = list(TASKGROUP._tasks)
    for task in tasks:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            print("Client task cancelled")

    RUNNING = False

async def run_conversation(addr, users):
    while True:
        await asyncio.sleep(2 + random.randrange(8))
        test_user = random.choice(users)
        tester = Conversation(addr, test_user)
        await tester.run()

async def main(server_address):
    global TASKGROUP, RUNNING
    RUNNING = True
    non_activated_users = [ {"username":"test",         "password":"test"},
                            {"username":"theviper",     "password":";(bQ,*0.zW/y@g"},
                            {"username":"kenny",        "password":"Hmmhmpf hmm hmmmpf"},
                            {"username":"rambro",       "password":"_a863TC69lN5JaUd"},
                            {"username":"r2d2",         "password":"not_a_bot"}]
    activated_users = [ {"username":"godfather",     "password":"TheLordWatchingYou"},
                        {"username":"julio_mafioso", "password":"abc123"},
                        {"username":"antonio",       "password":"G7Amj28mOwZ05Fnt2"},
                        {"username":"trinity",       "password":"D0dg3Th1s"},
                        {"username":"mario",         "password":"Its@ME3ee"}]
    async with asyncio.TaskGroup() as taskgroup:
        TASKGROUP = taskgroup
        taskgroup.create_task(run_conversation(server_address, non_activated_users))
        await asyncio.sleep(5)
        taskgroup.create_task(run_conversation(server_address, activated_users))
        await asyncio.sleep(5)
        taskgroup.create_task(run_conversation(server_address, [{}]))
    
if(__name__ == "__main__"):
    server_address = ("mitm", 42069)
    asyncio.run(main(server_address))
