#!/usr/bin/python3
import util
import json
import secrets
from sessions import Session


def load_json(msg):
    try:
        return json.loads(msg)
    except:
        raise util.ChallengeException("Unable to parse JSON: " + str(msg))


class ServerHandler(util.MessageHandler):
    def __init__(self, reader, writer, SESSIONS):
        super().__init__(reader, writer, "[SERVER]")
        self.SESSIONS = SESSIONS

    def parse_message(self, message_type):
        self.handler = None

        if self.session_id not in self.SESSIONS:
            if(self.session_id != b'\0' * util.SESSION_ID_LEN):
                raise util.ChallengeException("Invalid session ID")

            self.session_id = secrets.token_bytes(util.SESSION_ID_LEN)
            while self.session_id in self.SESSIONS:
                self.session_id = secrets.token_bytes(util.SESSION_ID_LEN)

            self.SESSIONS[self.session_id] = Session(self.session_id)

        messagetypes = [
                None, self.hello,    self.login,            self.logout,
                self.activatekey,    self.launcherversion,  self.connectivitycheck,
                self.listuserparams, self.changeuserparams, self.lastpatchnotes]

        if message_type in range(1, 10):
            self.handler = messagetypes[message_type]
        else:
            raise util.ChallengeException("Unknown message type")

    async def hello(self, msg, session):
        session.start()
        await self.send_message(1, "Hello")
        return True

    async def login(self, msg, session):
        session.check_state("disconnected")
        response = session.connect(msg)

        await self.send_message(2, json.dumps(response))
        return True

    async def logout(self, msg, session):
        session.check_state("connected")
        session.disconnect()
        await self.send_message(3, "Logout successful")
        return True

    async def activatekey(self, msg, session):
        session.check_state("connected")
        if message == ACTIVATION_KEY:
            session.activate()
            await self.send_message(4, "Activation successful")
            return True
        else:
            await self.send_message(0, "Invalid licence key")
            return False

    async def launcherversion(self, msg, session):
        session.check_state("disconnected")
        await self.send_message(5, "1.3.37")
        return True

    async def connectivitycheck(self, msg, session):
        session.check_state("disconnected")
        await self.send_message(6, msg)
        return True

    async def listuserparams(self, msg, session):
        session.check_state("connected")
        data = load_json(msg)
        response = session.dump_info(data)

        await self.send_message(7, json.dumps(response))
        return True

    async def changeuserparams(self, msg, session):
        session.check_state("connected")
        request = load_json(msg)

        if "password" in request:
            session.update_password(request["password"])
            await self.send_message(8, "Successfully updated user info: " +json.dumps(request))

        return True

    async def lastpatchnotes(self, msg, session):
        session.check_state("activated")
        request = load_json(msg)

        if "ts" not in request:
            await self.send_message(0, "'ts' field missing in request")
            return False
        if(type(request["ts"]) != int):
            await self.send_message(0, "'ts' field must be an integer")
            return False

        patchnotes = [
            {"ts":1691691690, "note":"Initial release"},
            {"ts":1693371000, "note":"Added clouds to the sky"},
            {"ts":1694291000, "note":"Removed cheatcodes because players kept abusing them"},
            {"ts":1694760000, "note":"Fixed bug where AI would stop playing and commit suicide"},
            #TODO: make more
                      ]
        response = []

        for patchnote in patchnotes:
            if(patchnote["ts"] >= request["ts"]):
                response.append(patchnote)

        await self.send_message(9, json.dumps(response))
        return True
