#!/usr/bin/python3
import time
import asyncio
from Crypto.Cipher import AES


SESSION_ID_LEN = 8
MSG_TYPE_LEN = 1
PAYLOAD_LEN_LEN = 2
HEADERS_LEN = SESSION_ID_LEN + MSG_TYPE_LEN + PAYLOAD_LEN_LEN

class ChallengeException(Exception):
    pass

class MessageHandler:
    def __init__(self, reader, writer, logprefix):
        self.reader = reader
        self.writer = writer
        self.logprefix = logprefix
        self.null_session = b'\0' * SESSION_ID_LEN
        self.session_id = self.null_session
        self.encryption_key = bytes.fromhex("9189cbcedb369ff590f25d52e00a8fa5")
        self.decryption_key = bytes.fromhex("d6fd67226abf32b8be423609f068284a")

    def log(self, message):
        print(self.logprefix + " " + message)

    def craft_message(self, msgtype, text):
        if(type(text) == str):
            encoded = text.encode()
        else:
            encoded = text

        if msgtype == 0:
            return b'\0' + encoded

        if not self.session_id:
            message = self.null_session + bytes([msgtype]) + len(encoded).to_bytes(2, "little") + encoded
        else:
            message = self.session_id + bytes([msgtype]) + len(encoded).to_bytes(2, "little") + encoded
        return b'\x01' + encrypt(message, self.encryption_key)

    async def send_message(self, msgtype, message):
        try:
            message = self.craft_message(msgtype, message)
            self.log(">>> Sending: " + str(message))
            self.writer.write(message)
            await self.writer.drain()
            return True
        except:
            return False

    async def receive_message(self):
        if not (received := (await self.reader.read(1024))):
            self.log("[-] Connection closed")
            return False

        try:
            message_type = self.parse_message_header(received)
            if message_type == 0:
                self.log("[-] Received error: " + str(self.payload))
                return False

            self.parse_message(message_type)
            self.log("<<< Received: " + str(self.payload))
            return True
        except ChallengeException as e:
            await self.send_message(0, str(e))
            return False

    async def handle(self, session = None):
        if not self.handler:
            return True

        try:
            return (await self.handler(self.payload, session))
        except ChallengeException as e:
            await self.send_message(0, str(e))
            return False

    def parse_message_header(self, message):
        if(len(message) < 1):
            raise ChallengeException("Status code required")
        status_code = message[0]
        self.session_id = b'\0' * SESSION_ID_LEN
        if(status_code == 0):
            self.payload = message[1:]
            return 0
        if(status_code != 1):
            raise ChallengeException("Invalid status code")
        encrypted = message[1:]
        if(len(encrypted) % 16 != 0):
            raise ChallengeException("Incomplete message")
        msg = decrypt(encrypted, self.decryption_key)

        if(len(msg) < HEADERS_LEN):
            raise ChallengeException("Message headers incomplete")

        self.session_id = msg[:SESSION_ID_LEN]
        message_type = msg[SESSION_ID_LEN]
        payload_len = msg[HEADERS_LEN - PAYLOAD_LEN_LEN:HEADERS_LEN]
        payload_len = int.from_bytes(payload_len, "little")

        if(payload_len > 1024 - HEADERS_LEN):
            raise ChallengeException("Payload length cannot be higher than " + str(1024 - HEADERS_LEN))
            
        if(len(msg) < HEADERS_LEN + payload_len):
            raise ChallengeException("Message is shorter than the announced " + str(payload_len) + " bytes")

        self.payload = msg[HEADERS_LEN:HEADERS_LEN + payload_len]

        return message_type


def verify_padding(msg):
    for i in reversed(range(len(msg))):
        if(msg[i] == 0x80):
            return True
        elif(msg[i] != 0):
            return False
    return False

def unpad(msg):
    if not verify_padding(msg):
        raise ChallengeException("Invalid ISO 7816-4 padding")

    for i in reversed(range(len(msg))):
        if(msg[i] == 0x80):
            return msg[:i]

def pad(msg):
    padlen = 16 - (len(msg) % 16)
    padding = b'\x80' + (b'\0' * (padlen - 1))
    return msg + padding

def encrypt(msg, key):
    cipher = AES.new(key, AES.MODE_CBC)
    return cipher.iv + cipher.encrypt(pad(msg))

def decrypt(msg, key):
    IV = msg[:16]
    msg = msg[16:]
    try:
        cipher = AES.new(key, AES.MODE_CBC, IV=IV)
    except ValueError as e:
        raise ChallengeException(str(e))
    plaintext = cipher.decrypt(msg)

    return unpad(plaintext)
